/***********************
 * Generate test graphs
 ***********************/

#include "defs.h"

#define TWOto16 65536

/* Definitions of utilities */
static long randi();
static void edge_add();
static long edge_in();
static unsigned long hash_edge();

/* large data structure for edges in graph */
static long edge_present[MAX_EDGE];

/* Length of MAX_EDGE*32 */
static long ep_length;

/* Generate a random graph (edge array) with v vertices and e edges.  It
 * will have a negative cycle of length cycle iff cycle >= 1.  Stores a
 * copy of the graph in g.
 * input: v, e, cycle (long), g (array of edges)
 * output: none
 * modifies: g
 * requires: g has size at least e+1; v < e <= v(v-1), cycle <= v
 */
void generate_graph(v,e,cycle,g)
long v,e,cycle;
edge *g;
{
  long i,j;
  unsigned long temp;
  long from, to;			/* temporary vertices */
  long origin;				/* starting vertex for cycle */
  long cycle_vertex[MAX_VERTEX];	/* which vertices are in cycle */
  long weight_adjust[MAX_VERTEX];	/* adjustment for each vertex */

  /* initialize ep_length */
  ep_length = 5;
  temp = MAX_EDGE;
  for (temp = MAX_EDGE-1; temp > 0; temp = temp >> 1){
    ep_length++;
  }

  /* initialize matrix of edges not allowed to add */
  for (i = 0; i < MAX_EDGE; i++){
    edge_present[i] = 0;
  }

  /* create negative cycle if one exists */
  if (cycle >= 1){
    cycle_vertex[0] = 0;
    for (i = 1; i < v; i++){		/* randomly permute vertices */
      j = randi(i+1);
      cycle_vertex[i] = cycle_vertex[j];/* note we define value at i here */
      cycle_vertex[j] = i;		/* works even when j==i */
    }
    for (i = 0; i < cycle-1; i++){	/* make cycle */
      g[i].from = cycle_vertex[i];
      g[i].to = cycle_vertex[i+1];
      g[i].weight = 0;
      edge_add(cycle_vertex[i],cycle_vertex[i+1]);
    }
    g[i].from = cycle_vertex[i];
    g[i].to = cycle_vertex[0];
    g[i].weight = -1;
    edge_add(cycle_vertex[i],cycle_vertex[0]);
    i = cycle;				/* i++ would also work here */
  }
  else{
    i = 0;				/* no edges added */
  }

  /* add remaining edges */
  /* i equals number of edges in graph so far */
  /* this could be slow for a dense graph */
  while (i < e){
    from = randi(v);
    to = randi(v);
    if (! edge_in(from,to)){
      g[i].from = from;
      g[i].to = to;
      g[i].weight = randi(TWOto16) + 1;	/* make sure positive */
      edge_add(from,to);
      i++;
    }
  }

  /* make edge weights random preserving cycle weights */
  for (i = 0; i < v; i++){
    weight_adjust[i] = randi(2*TWOto16)-TWOto16;
  }
  for (i = 0; i < e; i++){
    g[i].weight += weight_adjust[g[i].from];
    g[i].weight -= weight_adjust[g[i].to];
  }

  /* randomly permute edges */
  for (i = 1; i < e; i++){
    graph_edge_swap(g,i,randi(i+1));
  }
}

/*
 * Utilities
 */

/* Generate a random number between 0 and i-1.
 * Doesn't quite do this uniformly except for powers of 2,
 * but should be close enough.  Uses random().
 * input: i (long)
 * output: (long)
 */
static long randi(i)
long i;
{
  return(random() % i);
}

/* Determine if an edge is already present in the graph.
 * input: u,v [vertices] (long)
 * output: 0 if not present; 1 if present (long)
 *         (may output 1 if another edge with same hash is present)
 * global variable: edge_present
 * modifies: nothing
 */
static long edge_in(u,v)
long u,v;
{
  long hash;

  hash = hash_edge(u,v);
  return(edge_present[hash/32] & ((long) 1 << (hash % 32)));
}

/* Add an edge to a graph.
 * input: u,v [vertices] (long)
 * output: none
 * global variable: edge_present
 * modifies: edge_present
 */
static void edge_add(u,v)
long u,v;
{
  long hash;

  hash = hash_edge(u,v);
  edge_present[hash/32] |= ((long) 1 << (hash % 32));
}

/* Hash an edge.
 * input: u,v [vertices] (long)
 * output: (unsigned long)
 * global variable: ep_length
 * modifies: nothing
 */
static unsigned long hash_edge(u,v)
long u,v;
{
  unsigned long A;
  unsigned long temp;

  A = 2654435769;
  temp = ((u*A + v)*A);
  return(temp >> (32 - ep_length + 1));
}
