/*****************************************************************************
 * Name(s): George Madrid                                                    *
 * Email address: gamadrid@athena.mit.edu                                    *
 * Asymptotic worst-case running time: O(V + E)                              *
 * Brief description:                                                        *
 *     There are several steps involved:                                     *
 *     1)  Create an adjacency list representation of the graph.  O(E)       *
 *     2)  Depth first search to identify back edges.  O(V + E)              *
 *     3)  Remove back edges.  O(E)                                          *
 *     4)  Topologically sort the resultant DAG.  O(V + E)                   *
 *     5)  Find shortest path values for unconnected components of the graph.*
 *         O(V + E)                                                          *
 *     6)  See if the back edges will relax.  If so, continue.               *
 *         Otherwise, return NULL.  O(E)                                     *
 *     7)  Find shortest path from the relaxed back edge by relaxation.      *
 *         and print out the shortest path.                                  *
 *                                                                           *
 *     Unfortunately, this algorithms didn't quite work due to a theoretical *
 *     problem.  It is possible for a negative weight cycle to contain two   *
 *     back edges (when both back edges complete a zero weight cycle).       *
 *     Unfortunately, this breaks my algorithm.                              *
 *                                                                           *
 *     I am currently trying to find a work-around, because this algorithm is*
 *     very fast except for that case.  If I can find a work-around, then I  *
 *     will send it along just in case y'all are interested.  With all the   *
 *     rest of the end-of-the-term stuff, I just don't really have time.     *
 *     Sorry.                                                                *
 *                                                                           *
 *                                                                           *
 *     Oh well, back to the drawing board, and maybe I'll get some           *
 *     originality points.  :-)                                              *
 *                                                                           *
 *****************************************************************************/

#include "defs.h"  /* definition of edge, etc. */

/* ---------------------------------------------------------------------- */

/* Type declarations, etc... */

typedef long  wgt;
typedef long vertex;
typedef long  distance;
typedef long edge_num;

typedef enum { white = 0, black, grey } color;
typedef enum { forward = 0, back, tree} /* the cross-forward distinction is */
                           penguin;     /* unimportant */

/* Stuff for the adjacency list representation */
typedef struct {
    vertex from;  /* This is redundant.  Oh well. */
    vertex to;
    wgt weight;
    penguin characteristic;
    edge_num next; 
} myedge_struct, *myedge;

typedef struct {
    color  color;
    edge_num first;
    vertex toponext;		/* The next vertex topologically */
    wgt distance;		/* distance from "source"        */
    edge_num pred;              /* Nearest predecessor           */
} myvertex_struct, *myvertex;

/* ---------------------------------------------------------------------- */

/* Macros for use with the global graph adjacency list operator */

/* Vertices */
#define vertex_get_first(v)        (ernie[v].first)
#define vertex_set_first(v, f)     (ernie[v].first = (f))
#define vertex_get_color(v)        (ernie[v].color)
#define vertex_set_color(v, c)     (ernie[v].color = (c))
#define vertex_get_toponext(v)     (ernie[v].toponext)
#define vertex_set_toponext(v, n)  (ernie[v].toponext = (n))
#define vertex_get_distance(v)     (ernie[v].distance)
#define vertex_set_distance(v, n)  (ernie[v].distance = (n))
#define vertex_get_pred(v)         (ernie[v].pred)
#define vertex_set_pred(v, p)      (ernie[v].pred = (p))

/* Edges */
#define edge_get_from(e)           (bert[e].from)
#define edge_get_to(e)             (bert[e].to)
#define edge_get_weight(e)         (bert[e].weight)
#define edge_get_characteristic(e) (bert[e].characteristic)
#define edge_get_next(e)           (bert[e].next)

#define edge_set_from(e, f)           (bert[e].from = (f))
#define edge_set_to(e, t)             (bert[e].to = (t))
#define edge_set_weight(e, w)         (bert[e].weight = (w))
#define edge_set_characteristic(e, c) (bert[e].characteristic = (c))
#define edge_set_next(e, n)           (bert[e].next = (n))

#define string_for_color(c) \
    ((c == white) ? "white" : ((c == grey) ? "grey" : "black"))
#define string_for_characteristic(c) \
    ((c == forward) ? "forward" : ((c == back) ? "back" : "tree"))

#define make_first_topologically(i) \
    (vertex_set_toponext((i), grover), grover = (i))

/* ---------------------------------------------------------------------- */

/* Constants */
#define NULL_EDGE -1
#define NULL_VERTEX -1
#define INFINITY 0x7FFFFFF0

/* ---------------------------------------------------------------------- */

/* Global declarations */

/* bert is an array of edges because he is so edgy. */
myedge_struct bert[MAX_EDGE];
edge_num num_edges = NULL_EDGE;

/* ernie is an array of vertices because he has a big node.  :-) */
myvertex_struct ernie[MAX_VERTEX];
vertex num_vertices = NULL_VERTEX;

/* bigbird is a pointer to the list of back edges */
edge_num bigbird = NULL_EDGE;

/* grover is a pointer to the first vertex topologically */
vertex grover = NULL_VERTEX;

myedge kermit[MAX_EDGE];

/* ---------------------------------------------------------------------- */

/* Local function declarations */
int relax();
void dfs_visit();
void dfs();
int find_negative_cycle();
void construct_negative_list();
void create_adjacency_list();
void shortest_paths();
void shortest_path_from_vertex();
void unparse_adjacency_list();
void unparse_grover();
void unparse_bigbird();
void check_kermit();

/* ---------------------------------------------------------------------- */
/*
  void create_adjacency_list(edge **edgelist, int nv)
  
  Effects:  Create an adjacency list representation of the graph in g, putting
            the graph into the global adjacency list.  nv is the number of
	    vertices in the graph. 
  Modifies: The global graph
*/

void create_adjacency_list(edgelist, nv)
    edge **edgelist;
    long nv;
{
    int i;
    edge_num edge_number = 0;
    edge *this_edge = *edgelist;

    /* Initialize the vertex list */
    for (i = 0; i < nv; i++) {
	vertex_set_first(i, NULL_EDGE);
	vertex_set_color(i, white);
	vertex_set_distance(i, INFINITY); /* Set this now */
	vertex_set_pred(i, NULL_EDGE);
    }

    /* Do this until you hit a NULL edge. */
    while (this_edge = edgelist[edge_number]) {
	/* Now do the edge initializations */
	edge_set_from(edge_number, this_edge->from);
	edge_set_to(edge_number, this_edge->to);
	edge_set_weight(edge_number, this_edge->weight);
	edge_set_characteristic(edge_number, forward);

	/* Add to the adjacency list */
	edge_set_next(edge_number, vertex_get_first(this_edge->from));
	vertex_set_first(this_edge->from, edge_number);

	edge_number++;
    }
    num_edges = edge_number;
    num_vertices = nv;
}

/* ---------------------------------------------------------------------- */

void unparse_adjacency_list()
/*
  void unparse_adjacency_list();

  Requires: the global adjacency list is somewhere.
  Effects:  Print a human readable version of the graph stored in the global
            adjacency list.
 */
{
    vertex i;
    edge_num curredge;

    printf("\n\nAdjacency list rep:\n");
    for(i = 0; i < num_vertices; i++) {
	curredge = ernie[i].first;
	printf("From vertex %6d(%s)[pred %d %6ld]:  \n",
	       i, 
	       string_for_color(vertex_get_color(i)),
	       ((vertex_get_pred(i) != NULL_EDGE) ? 
		vertex_get_pred(i) :
		-1),
	       vertex_get_distance(i));

	while(curredge != NULL_EDGE) {
	    printf("    %6d: [%6d %6d](%8ld) %s  ->  %6d\n",
		   curredge,
		   edge_get_from(curredge),
		   edge_get_to(curredge),
		   edge_get_weight(curredge),
		   string_for_characteristic(edge_get_characteristic(curredge)),
		   edge_get_next(curredge));
	    curredge = edge_get_next(curredge);
	}
	printf("\n");
    }
}

void unparse_bigbird()
{
    int i = 0;
    edge_num curredge;
    
    printf("\nBACK EDGES:\n");
    curredge = bigbird;
    while(curredge != NULL_EDGE) {
	printf("    %6d: [%6d %6d](%8ld) %s  ->  %6d\n",
	       curredge,
	       edge_get_from(curredge),
	       edge_get_to(curredge),
	       edge_get_weight(curredge),
	       string_for_characteristic(edge_get_characteristic(curredge)),
	       edge_get_next(curredge));
	curredge = edge_get_next(curredge);
	i++;
    }
    printf("There were %d back edges.\n\n", i);
}

void unparse_grover()
{
    int i = 0;
    vertex currvert;

    printf("TOPOLOGICALLY:\n");
    currvert = grover;
    while(currvert != NULL_VERTEX) {
	printf("%6d->", currvert);
	currvert = vertex_get_toponext(currvert);
	i++;
    }
    printf("END of %d vertices.\n", i);
}

/* ---------------------------------------------------------------------- */

void dfs()
{
  vertex i;

  /* Code stolen from CLR page 478. */
  /* Vertex color initialized when created. */
  /* I'm not concerned about predecessors or discover times. */

  for(i = 0; i < num_vertices; i++) {
    if (vertex_get_color(i) == white) {
      if (DEBUG) {
	printf("DFS NEW SOURCE:  %d\n", i);
      }
      dfs_visit(i);
    }
  }
}

void dfs_visit(i)
     vertex i;     
{
    edge_num curredge;
    edge_num prev;		/* The previous edge in the adjacency list */

    vertex_set_color(i, grey);

    curredge = vertex_get_first(i);
    prev = NULL_EDGE;

    while (curredge != NULL_EDGE) {
	if (vertex_get_color(edge_get_to(curredge)) == white) {
	    edge_set_characteristic(curredge, tree);
	    dfs_visit(edge_get_to(curredge));
	    prev = curredge;
	    curredge = edge_get_next(curredge);
	}
	else if (vertex_get_color(edge_get_to(curredge)) == grey) {
	    edge_num next;

	    edge_set_characteristic(curredge, back);

	    /* Now put it in bigbird */
	    if (prev == NULL_EDGE) {
		/* Remove from beginning of list */
		vertex_set_first(i, edge_get_next(curredge));
	    }
	    else {
		/* Remove from middle of the list */
		edge_set_next(prev, edge_get_next(curredge));
	    }

	    /* Don't reset prev, because we just removed this edge */
	    next = edge_get_next(curredge);
	    edge_set_next(curredge, bigbird);
	    bigbird = curredge;
	    curredge = next;
	}
	else {
	    edge_set_characteristic(curredge, forward);
	    prev = curredge;
	    curredge = edge_get_next(curredge);
	}
    }
    vertex_set_color(i, black);
    make_first_topologically(i);
}

/* ---------------------------------------------------------------------- */

void shortest_paths()
{
    vertex i;
    edge_num curredge;

    for (i = grover; i != NULL_VERTEX; i = vertex_get_toponext(i)) {
	curredge = vertex_get_first(i);
	if (vertex_get_distance(i) == INFINITY) {
	  vertex_set_pred(i, NULL_EDGE);
	  vertex_set_distance(i, 0);
	}
	while (curredge != NULL_EDGE) {
	  relax(curredge);
	  curredge = edge_get_next(curredge);
	}
    }
}

void shortest_path_from_vertex(i)
     vertex i;
{
  vertex v;
  edge_num curredge;

  if (DEBUG) {
      printf("shortest_path_from_vertex(%ld);\n", i);
  }

  /* Reset all of these values */
  for (v = 0; v < num_vertices; v++) {
    vertex_set_distance(v, INFINITY);
  }

  /* Do this garbage again */
  for (v = i; v != NULL_VERTEX; v = vertex_get_toponext(v)) {
    curredge = vertex_get_first(v);
    if (vertex_get_distance(v) == INFINITY) {
      vertex_set_pred(v, NULL_EDGE);
      vertex_set_distance(v, 0);
    }
    while (curredge != NULL_EDGE) {
      relax(curredge);
      curredge = edge_get_next(curredge);
    }
  }
  printf("\n\nAdjacency list after relaxing again\n");
  unparse_adjacency_list();
}

int relax(e)
     edge_num e;
{
  if (DEBUG) {
    printf("Relaxing (%6d, %6d).  From: %11ld  Old: %11ld  New:%11ld    ", 
	   edge_get_from(e),
	   edge_get_to(e),
	   vertex_get_distance(edge_get_from(e)),
	   vertex_get_distance(edge_get_to(e)),
	   vertex_get_distance(edge_get_from(e)) + edge_get_weight(e));
  }
  
  if (vertex_get_distance(edge_get_to(e)) >
      (vertex_get_distance(edge_get_from(e)) +
       edge_get_weight(e))) {
    if (DEBUG) {
      printf("    relaxed.(pred=%d(%d))\n",
	     edge_get_to(e),
	     e);
    }
    vertex_set_distance(edge_get_to(e),
			vertex_get_distance(edge_get_from(e)) +
			edge_get_weight(e));
    vertex_set_pred(edge_get_to(e), e);
    return(1);
  } else {
    if (DEBUG) {
      printf("NOT relaxed.\n");
    }
    return(0);
  }
}

/* ---------------------------------------------------------------------- */

int find_negative_cycle()
{
  edge_num i;

  for (i = bigbird; i != NULL_EDGE; i = edge_get_next(i)) {
    if (relax(i)) {
      shortest_path_from_vertex(edge_get_to(i));
      construct_negative_list(i);
      return(1);
    }
  }
  printf("\nAfter relaxing back edges.\n");
  unparse_adjacency_list();
  return(0);
}

void construct_negative_list(i)
    edge_num i;
{
    int k = 0;
    edge_num j;
    
    for (j = i; edge_get_from(j) != edge_get_to(i); 
	 j = vertex_get_pred(edge_get_from(j))) {
      if (DEBUG) {
	printf("construct_negative_list:  kermit[%6d] = [%6d, %6d]\n",
	       k, edge_get_from(j), edge_get_to(j));
      }
      kermit[k++] = bert + j;
    }
    kermit[k++] = bert + j;
    kermit[k] = NULL;
}

void check_kermit()
{
  long i;
  long total;

  for(i = 0, total = 0; kermit[i] == NULL; i++) {
      total += kermit[i]->weight;
  }
  
  printf("%ld edges at a weight of %ld\n", i, total);
}

/* ---------------------------------------------------------------------- */
/* Find a negative-weight cycle.
 * input: g (array of pointers to edges), nv [number of vertices] (short)
 * output: array of pointers (terminated by NULL) to edges of a negative cycle
 *  or NULL if there is no negative cycle
 * modifies: nothing
 * requires: g is terminated by NULL; the only vertices in g are in the range
 *           [0,nv-1] 
 */
edge** negative_cycle(g,nv)
 edge **g;	/* pointer to the edge list */
 long nv;	/* number of vertices in the graph */
{
    /* 1.  Create an adjacency list representation of the graph.  O(E) */
    create_adjacency_list(g, nv);

    /* 2.  Depth first search to identify back edges.  O(V + E)        */
    /* 3.  Remove back edges.  O(E)                                    */
    /* 4.  Topologically sort the resultant DAG.  O(V + E)             */
    dfs();
    
    /* 5.  Find shortest path values for unconnected components        */
    /*     of the graph.  O(V + E)                                     */
    shortest_paths();
    
    if (DEBUG) {
	unparse_adjacency_list();
	unparse_bigbird();
	unparse_grover();
    }

    /* 6.  See if the back edges will relax.  If so, return the        */
    /*     negative weight cycle.  Otherwise, return NULL.  O(E)       */
    if (find_negative_cycle()) {
      check_kermit();
      return((edge **)kermit);
    }
    else {
	return(NULL);
    }
}

/* ---------------------------------------------------------------------- */

/*

  Problems:



*/
