/*****************************************************************************
 *              AUTHOR HEADER                                                *
 * (Replace with your own name and address)                                  *
 *                                                                           *
 * Name(s): John Leo                                                         *
 * Email address: leo@theory.lcs.mit.edu                                     *
 * Asymptotic worst-case running time: O(VE^2)                               *
 * Brief description:                                                        *
 *  Uses Bellman-Ford to determine of there is a negative weight cycle, then *
 *  removes an edge and tests again.                                         *
 *****************************************************************************/

#include "defs.h"  /* definition of edge, etc. */

/**** declarations of functions below ****/
long bellman_ford();
void relax();

/**************************
 **** negative_cycle() ****
 **************************/

/* Find a negative-weight cycle.
 * input: g (array of pointers to edges), nv [number of vertices] (long)
 * output: array of pointers (terminated by NULL) to edges of a negative cycle
 *  or NULL if there is no negative cycle
 * modifies: nothing
 * requires: g is terminated by NULL; the only vertices in g are in the range
 *           [0,nv-1] 
 */
edge** negative_cycle(g,nv)
 edge **g;	/* pointer to the edge list */
 long nv;	/* number of vertices in the graph */
{
  edge *graph[MAX_EDGE];		/* edge list to work with */
  long pos;				/* position in graph */
  long ne;				/* number of edges */

  ne = el_copy(graph,g);		/* copy g to graph and get size */
  if (bellman_ford(graph,nv,ne)){
    return(NULL);			/* no negative cycle */
  }
  pos = ne-1;				/* last edge in graph */
  while (pos >= 0){			/* at least one edge left */
    edge_swap(graph,pos,ne-1);		/* swap current with last edge */
    if (! bellman_ford(graph,nv,ne-1)){	/* edge not needed */
      graph[--ne] = NULL;			/* delete edge */
    }
    pos--;				/* look at next edge */
  }
  graph[ne] = NULL;			/* add NULL termination */
  return(graph);
}

/*****************************
 **** Bellman-Ford Algorithm *
 *****************************/

/* Runs Bellman-Ford algorithm (CLR, page 532) for the purpose of finding
 *  whether there is a negative-weight cycle.  Note there is no source.
 *  Returns 1 if no cycle is found; 0 if a cycle is found. 
 *  Edge list need not be NULL-terminated because the number
 *  of edges is also given as an input.
 *  Starts with distance 0 for each vertex instead of infinity.
 *  Also the predecessor information is not kept.
 * input: graph (array of pointers to edges), nv [number of vertices] (long),
 *        ne [number of edges] (long)
 * output: 0 if negative-weight cycle found, 1 otherwise (long)
 * modifies: nothing
 * requires: the only vertices in the edge list are in the range [0,nv-1]
 */
long bellman_ford(graph,nv,ne)
edge **graph;
long nv,ne;
{
  long d[MAX_VERTEX];		/* distance array */
  long i,j;

  /* Initialize-Single-Source (CLR page 520).
   * However there is no source here, and all initial distances are zero.
   * Also predecessor info is not kept. */
   for (i = 0; i < nv; i++){
     d[i] = 0;
   }

  /* main loop */
  for (i = 0; i < nv-1; i++){
    for (j = 0; j < ne; j++){
      relax(graph[j]->from,graph[j]->to,graph[j]->weight,d);
    }
  }

  /* check for negative cycle */
  for (j = 0; j < ne; j++){
    if (d[graph[j]->to] > d[graph[j]->from] + graph[j]->weight){
      return(0);
    }
  }
  return(1);
}

/* Relax (CLR page 520).  However predecessor array is omitted.
 * input: u [from vertex] (long), v [to vertex] (long),
 *        w [weight of (u,v)] (long), d [distance array] (array of long)
 * output: none
 * modifies: *d
 */
void relax(u,v,w,d)
long u,v;
long w,*d;
{
  if (d[v] > d[u] + w){
    d[v] = d[u] + w;
  }
}

/*******************************
 * Utilities defined in test.c *
 *******************************/

/* Copy one edge list to another.
 * input: g1 (array of pointers to edges), g2 (array of pointers to edges)
 * output: number of edges (long)
 * modifies: g1
 * requires: g1 be at least as big as g2; g2 be terminated by NULL
 */
/* long el_copy(g1,g2) */

/* Swap two edges in an edge list.
 * input: graph (array of pointers to edges), pos1, pos2 [positions of edges
 *       to be swapped] (long)
 * output: none
 * modifies: *graph
 */
/* void edge_swap(graph,pos1,pos2) */
